# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Implementation of RAI Vision Insights API."""

from responsibleai_vision.rai_vision_insights.rai_vision_insights import \
    RAIVisionInsights

__all__ = ['RAIVisionInsights']
